#include <stdio.h>
#include <math.h>

#ifdef DOS
#include <stdlib.h>
#include <conio.h>
#define MAXINT 32767
#endif

#ifdef UNIX

#include <sys/ioctl.h>
#include <unistd.h>
#include <stdlib.h>

#ifdef BSD
#include <termios.h>
#elif defined NEXT
#include <malloc.h>
#include <sgtty.h>
#else /* SysV */
#include <malloc.h>
#include <termio.h>
#endif /* ifdef BSD */
#endif /* UNIX */

#if defined(UNIX) && defined(HOTKEYS)

#ifdef BSD
struct termios raw;    /* characteristics for hotkeys */
struct termios noraw;  /* original characteristics */
#elif defined NEXT
struct sgttyb raw;     /* characteristics for hotkeys */
struct sgttyb noraw;   /* original characteristics */
#else /* the System V solution */
struct termio raw;    /* characteristics for hotkeys */
struct termio noraw;  /* original characteristics */
#endif

char onechar[1];      /* the buffer for the hotkey character */
#endif  /* if defined(UNIX) && defined(HOTKEYS) */

#ifdef UNIX
#define SIGINT 2
#elif defined DOS16
#include <stdlib.h>
#include <conio.h>
#else
#include <stdlib.h>
#include <conio.h>
#endif

#define SIZE 26          /* one larger than the number of nodes */
#define M 25             /* the number of nodes */
#define VALSPERLINE 5    /* the number of values printed per line */
#define L 1.05           /* 1.05 works */
#define buffsize 258     /* size minus 1 of the input buffer */

char outstr[buffsize];   /* output buffer size */
float f1[SIZE];          /* f1 register */
float f2[SIZE];          /* f2 register */
float x[SIZE];           /* intersection of input with top-down pat */
int learned[SIZE];       /* flags the f2 nodes that have templates */
float up[SIZE][SIZE];    /* bottom-up weights */
float down[SIZE][SIZE];  /* top-down weights */
FILE *data;              /* file for original data */
FILE *weights;           /* file for saving weights */
char datafilename[15];   /* copy of the data file name saved here */
char cmdfilename[15];    /* name of file to take extra commands from */
int  bufferend;          /* length of buffer input line */
int  bufferptr;          /* position of next character in buffer */
char buffer[buffsize];   /* holds contents of one input line */
int echo;                /* to control printing of input */
float rho;               /* vigilance parameter */
char ch;                 /* general purpose character variable */
int lineno;              /* counts lines for paging */
int copyflag;            /* the make a copy of i/o flag */
int pagesize;            /* size of a screen page */
FILE *copyfile;          /* pointer to the file copy */
int nread;               /* number of values read by sscanf */

#if defined(UNIX) && defined(HOTKEYS)

#ifdef BSD
void initraw()
{
 int i;
 tcgetattr(0,&noraw);
 raw.c_iflag = noraw.c_iflag;
 raw.c_oflag = noraw.c_oflag;
 raw.c_cflag = noraw.c_cflag;
 raw.c_lflag = noraw.c_lflag & ~ ECHO;
 raw.c_lflag = raw.c_lflag & ~ ICANON;
 for (i=0;i<NCCS;i++) raw.c_cc[i] = noraw.c_cc[i];
 raw.c_cc[VMIN] = 1;
 raw.c_cc[VTIME] = 0;
 raw.c_ispeed = noraw.c_ispeed;
 raw.c_ospeed = noraw.c_ospeed;
}

void setraw()
{
 tcsetattr(0,TCSANOW,&raw);
}

void setnoraw()
{
 tcsetattr(0,TCSANOW,&noraw);
}

#elif defined NEXT
void initraw()
{
 ioctl(0,TIOCGETD,&noraw);
 raw.sg_ispeed = noraw.sg_ispeed;
 raw.sg_ospeed = noraw.sg_ospeed;
 raw.sg_erase = noraw.sg_erase;
 raw.sg_kill = noraw.sg_kill;
 raw.sg_flags = noraw.sg_flags & 0000040;   /* turn on raw mode */
}

void setraw()
{
 ioctl(0,TIOCSETD,&raw);
}

void setnoraw()
{
 ioctl(0,TIOCSETD,&noraw);
}

#else /* System V code */

void initraw()
{
 int i;
 ioctl(0,TCGETA,&noraw);
 raw.c_iflag = noraw.c_iflag;
 raw.c_oflag = noraw.c_oflag;
 raw.c_cflag = noraw.c_oflag;
 raw.c_lflag = noraw.c_lflag & ~ 0x02;   /* hot key bit */
 raw.c_lflag = raw.c_lflag & ~ 010;      /* echo bit */
 raw.c_line = noraw.c_line;
 for (i=0;i<=7;i++) raw.c_cc[i] = noraw.c_cc[i];
 raw.c_cc[4] = 1;
}

void setraw()
{
 ioctl(0,TCSETA,&raw);
}

void setnoraw()
{
 ioctl(0,TCSETA,&noraw);
}
#endif /* end of the System V code */

int getch()
{
 int ch;
 setraw();
 do ch = read(0,onechar,1); while (ch != 1);
 setnoraw();
 return(onechar[0]);
}
#endif  /* end of the if defined(UNIX) ... code */

int pg(char *str) /* paging and making a copy function */
{
char *ch3,action,cr;
int i;
int nbackspaces = 5;

action = 0;
ch3 = str;
while (*ch3 != '\0')
 {
  if (*ch3 == '\n')
   {
    putchar('\n');
    if (copyflag == 1) putc('\n',copyfile);
    lineno = lineno + 1;
    if (pagesize > 0 && lineno > 0 && lineno % pagesize == 0)
     {
      printf("More?");
#ifdef HOTKEYS
      action = getch();
#else
      action = getchar();
      if (action != '\n') cr = getchar();
#endif
      for(i=1;i<=nbackspaces;i++) putchar(8);
      for(i=1;i<=nbackspaces;i++) putchar(' ');
      for(i=1;i<=nbackspaces;i++) putchar(8);
      if (action == 'q') return(1);
      else if (action == 'c') lineno = -MAXINT;
#ifdef UNIX
      else if (action == '\n') lineno = lineno - 1;
#else
      else if (action == 13) lineno = lineno - 1;
#endif
      else if (action == 4) lineno = pagesize / 2;
      else lineno = 0;
     }
   }
  else {putchar(*ch3); if (copyflag == 1) putc(*ch3,copyfile); };
  ch3++;
 };
return(0);
}

int readch()
{ 
  int i, ch2;
  if(bufferptr > bufferend) 
    {
      ch2 = getc(data);
      if(ch2 == EOF) return(ch2);
      i = 0;
      while(ch2 != '\n' && i < (buffsize-2))
         {
           buffer[i] = ch2;
           i = i + 1;
           ch2 = getc(data);
         };
      if(i == (buffsize-2))
         {
           pg("line too long\n");
           exit(4);
         };
      buffer[i] = '\n';
      buffer[i+1] = '\0';
      bufferend = i;
      bufferptr = 0;
      if (echo == 1) pg(buffer);
     }
  ch2 = buffer[bufferptr];
  bufferptr = bufferptr + 1;
  return(ch2);
}

int readchar()        /* reads data in compressed format */
{ char ch2;
  ch2 = readch();
  do {
         switch (ch2) {
       case '\n':
       case ' ': ch2 = readch();
                 break;
       case '1': return(1);
                 break;
       case '0': return(0);
                 break;
       case '*': do {ch2 = readch();} while(ch2 != '\n');
                 break;
        default: sprintf(outstr,"unexpected character:%c\n",ch2);
                 pg(outstr);
                 return(-1);
                 break;
          }; /* end switch */
  } while (0 == 0);
}

char *extractnum()
{
 int i,j,k,length;
 char *strptr, *s;
 i = bufferptr;
 while (buffer[i] == ' ') i = i + 1;
 j = i;
 while (buffer[j] != '*' && buffer[j] != '\n' && buffer[j] != ' ') j = j + 1;
 bufferptr = j;
 length = j - i + 1;
 strptr = (char *) malloc(length);
 s = strptr;
 for (k=i;k<=bufferptr-1;k++) *s++ = buffer[k];
 *s = '\0';
 return(strptr);
}

void init()
{
  float randomup;
  int i,j;

  bufferend = 0;
  bufferptr = buffsize + 1;
  ch = ' ';
  copyflag = 0;
  lineno = 0;
  pagesize = 23;
  echo = 0;
  rho = 0.5;

  randomup = (float) L / ((float) (L - 1 + M));
  for (i=1;i<=M;i++)
      for (j=1;j<=M;j++) 
          up[i][j] = randomup;
  for (i=1;i<=M;i++)
      for (j=1;j<=M;j++)
          down[i][j] = 1.00;

  for (i=1;i<=M;i++) learned[i] = 0;
}

void identify()
{
  int i, j, place;
  float ratio, max;
  int res, icount, xcount;

/* count the number of 1 bits in the input */

  icount = 0;
  for (i=1;i<=M;i++)
      if (f1[i] == 1.0) icount = icount + 1;

/* activate the f2 nodes */

  for (i=1;i<=M;i++) f2[i] = 0;
  for (i=1;i<=M;i++)
      for (j=1;j<=M;j++)
           f2[i] = f2[i] + up[i][j] * f1[j];

/* print the f2 nodes */

   pg("\nf2 nodes:\n");
   for (i=1;i<=M;i++)
      {
       sprintf(outstr,"%5.2f",f2[i]);
       pg(outstr);
       if (i % VALSPERLINE == 0) pg("\n");
      };
   pg("\n");

/* check f2 nodes in order according to decreasing activation */

  res = 0;
  while (res == 0)
     {
        max = f2[1];
        place = 1;
        for (i=2;i<=M;i++)
           if (f2[i] > max)
              {
                max = f2[i];
                place = i;
              };
        if (learned[place] == 0)
           {
             for (i=1;i<=M;i++) x[i] = f1[i];
             res = 1;
             xcount = icount;
             sprintf(outstr,"checking on pattern %d; max = %f\n",place,max);
             pg(outstr);
             sprintf(outstr,"RESONANCE f2 node %d NEW PATTERN\n\n",place);
             pg(outstr);
           }
        else
           {
             sprintf(outstr,"checking on pattern %d; max = %f\n",place,max);
             pg(outstr);
             /* match templates in down weights against f1 */
             pg("\ndown weights:\n");
             for (i=1;i<=M;i++)
                 {
                   if (down[place][i] == 1.0 && f1[i] == 1.0) 
                      x[i] = 1;
                   else x[i] = 0;
                 };
             for (i=1;i<=M;i++)
                 if (i % VALSPERLINE == 0)
                    {sprintf(outstr,"%5.2f\n",x[i]); pg(outstr); }
                 else
                    {sprintf(outstr,"%5.2f",x[i]); pg(outstr); };
             pg("\n");
             xcount = 0;
             for (i=1;i<=M;i++)
                 if (x[i] == 1) xcount = xcount +1;
             ratio = (float) xcount / (float) icount;
             sprintf(outstr,"%d matches out of %d: %5.2f", xcount,icount,ratio);
             pg(outstr);
             if (ratio >= rho)
                {
                  sprintf(outstr," RESONANCE f2 node %d\n\n",place);
                  pg(outstr);
                  res = 1;
                }
             else
                {
                  pg(" fails\n\n");
                  f2[place] = - 1.0;
                };
           };
     };

/* mark the node as having a template */

     learned[place] = 1;

/* modify the top-down templates */

     for (i=1;i<=M;i++)
         if (x[i] == 1.0) down[place][i] = 1.0;
         else down[place][i] = 0.0;

/* modify the bottom-up templates */

     for (i=1;i<=M;i++)
         if (x[i] == 1.0)
            up[place][i] = (float) L / ((float)(L - 1 + xcount));
         else up[place][i] = 0.0;
}
  
void cmdloop()    /* read commands and process them */

{
 int finished, count;
 int i, j, temp, itemp;
 char string[128], *buf;

 finished = 0;
 do{
restart:
     if (data == stdin)
        {
         lineno = 0;
         pg("[?!*cdehipqru]? ");
        };
     ch = readch();
     while(ch == ' ' || ch == '\n') ch = readch();
         switch (ch) {
     case EOF : data = stdin;
                pg("taking commands from stdin now\n");
                bufferend = 0;  /* force it to read from stdin */
                bufferptr = buffsize + 1;  /* when readch is called */
                goto restart;
     case '?' : sprintf(outstr,"%5.2f\n",rho); pg(outstr);
                break;
     case '!' : count = 0;
                ch = readch();
                while (ch != '\n' && count <= 127)
                   {
                     string[count] = ch;
                     ch = readch();
                     count = count + 1;
                   };
                bufferptr = bufferptr - 1; /* ungetc(ch,data); */
                string[count] = '\0';
                system(string);
                break;
     case '*' : break;  /* * on a line is a comment */
     case 'c' : buf = extractnum();
                nread = sscanf(buf,"%d",&itemp);
                if (nread == 0) pg("copy flag not read\n");
                else
                   {
                    if (itemp == 1)
                       {
                        copyflag = 1;
                        copyfile = fopen("copy","w");
                       }
                    else if (itemp == 0)
                       {
                        if (copyflag == 1) fclose(copyfile);
                        copyflag = 0;
                       }
                    else pg("bad value for copy flag\n");
                   };
                break;
     case 'e' : buf = extractnum();
                nread = sscanf(buf,"%d",&echo);
                if (nread == 0) pg("echo flag not read\n");
                break;
     case 'h' : pg("? prints out the value of rho, the vigilance parameter\n");
                pg("! shell escape\n");
                pg("* * on a line is a comment\n");
                pg("c 1 turns on copy function, c 0 turns it off\n");
                pg("d <int> prints weights going down from unit <int>\n");
                pg("e 1 echos input, e 0 stops the echo\n");
                pg("h give this help listing\n");
                pg("i <filename> takes input from <filename>\n");
                pg("p <pattern> inputs a pattern\n");
                pg("q quits the program\n");
                pg("r <real> sets rho the vigilance parameter to <real>\n");
                pg("u <int> prints weights going up to unit <int>\n");
                break;
     case 'i' : ch = readch();
                while(ch == ' ') ch = readch();
                itemp = 0;
                while(ch != ' ' && ch != '\n' && itemp < 15)
                   {
                     cmdfilename[itemp] = ch;
                     itemp = itemp + 1;
                     ch = readch();
                   };
                cmdfilename[itemp] = '\0';
                if((data = fopen(cmdfilename,"r")) == (FILE *) NULL)
                   {
                     sprintf(outstr,"cannot open: %s\n",cmdfilename);
                     pg(outstr);
                     data = stdin;
                     pg("taking commands from stdin now\n");
                   }
                bufferend = 0;
                bufferptr = buffsize + 1;
                ch = ' ';
                goto restart;
                break;
     case 'p' : for (i=1;i<=M;i++) f1[i] = readchar();
                identify();
                break;
     case 'q' : finished = 1;
                break;
     case 'r' : buf = extractnum();
                nread = sscanf(buf,"%f",&rho);
                if (nread == 0) pg("rho not read\n");
                break;
     case 'u' : buf = extractnum();
                nread = sscanf(buf,"%d",&itemp);
                if (nread == 0) pg("missing node number\n");
                else
                   {
                    for (i=1;i<=M;i++)
                       if (i % VALSPERLINE == 0)
                          {
                           sprintf(outstr,"%5.2f\n",up[itemp][i]);
                           pg(outstr);
                          }
                       else
                          {
                           sprintf(outstr,"%5.2f",up[itemp][i]);
                           pg(outstr);
                          };
                    pg("\n");
                   };
                break;
     case 'd' : buf = extractnum();
                nread = sscanf(buf,"%d",&itemp);
                if (nread == 0) pg("missing node number\n");
                else
                   {
                    for (i=1;i<=M;i++)
                       if (i % VALSPERLINE == 0)
                          {
                           sprintf(outstr,"%5.2f\n",down[itemp][i]);
                           pg(outstr);
                          }
                       else
                          {
                           sprintf(outstr,"%5.2f",down[itemp][i]);
                           pg(outstr);
                          };
                    pg("\n");
                   };
                break;
     default  : pg("unrecognized command; skipping to next line\n");
                break;
         };
     do ch = readch(); while (ch != '\n');
  }while (finished == 0);
};

void main(int argc,char *argv[])
{
 char *fnamestr, *i;

 setbuf(stdout,NULL);  /* unbuffered output */
#if defined(UNIX) && defined(HOTKEYS)
 initraw();
#endif

 /* check for file argument, if any */

 if (argc == 1)
    {
      pg("missing data file name, stdin assumed\n");
      data = stdin;
      *datafilename = '\n';
    }
 else
    if((data = fopen(argv[1],"r")) == (FILE *) NULL)
       {
         sprintf(outstr,"cannot open: %s\n",argv[1]);
         pg(outstr);
         exit(1);
       }
    else /* make a copy of the file name in a global variable */
       {
         fnamestr = argv[1];
         i = datafilename;
         while(*fnamestr != '\0') *i++ = *fnamestr++;
       };
 init();
 cmdloop();
}
